/*
Copyright © Brian Buller <brian@bullercodeworks.com>

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.
*/
package widgets

type LayoutFlag int

// Alignment Flags
// The default should be fully centered
const (
	_              = LayoutFlag(iota)
	LFAlignHLeft   // 01
	LFAlignHRight  // 10
	LFAlignHCenter // 11

	LFAlignH = LayoutFlag(0x11)
)

const (
	_              = LayoutFlag(iota << 2)
	LFAlignVTop    // 0100
	LFAlignVBottom // 1000
	LFAlignVCenter // 1100

	LFAlignV = LayoutFlag(0x1100)
)

func (f LayoutFlag) Add(fl LayoutFlag)  { f |= fl }
func (f LayoutFlag) AlignH() LayoutFlag { return f & LFAlignH }
func (f LayoutFlag) AlignV() LayoutFlag { return f & LFAlignV }
func (f LayoutFlag) IsAlignH() bool     { return f.AlignH() > 0 }
func (f LayoutFlag) IsAlignV() bool     { return f.AlignV() > 0 }
func (f LayoutFlag) ClearAlignH() {
	f = f &^ LFAlignH
	f.Add(LFAlignHCenter)
}

func (f LayoutFlag) ClearAlignV() {
	f = f &^ LFAlignV
	f.Add(LFAlignVCenter)
}
